/* rexec.c
**
** check if a user has the permission to execute a command recipient.
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

#include "mrmacs.h"
#include "inline-funcs.h"
#include "prefs.h"
#include "tools.h"

#include "rexec.h"

/* Was will ich hier eigentlich machen?
**   Am besten ist es, nur solche Programme zur Ausfhrung zuzulassen,
**   die in einer `trusted list' stehen. Damit eine solche `trusted list'
**   nicht von jedem beliebigen erzeugt, verndert sonstwie manipuliert
**   werden kann, mu sie unabhngig von der Konfiguration angelegt und
**   verwaltet werden. Dazu darf sie nicht in der normalen Konfigurations-
**   datei stehen; sie wird stattdessen direkt in den Code kompiliert;
**   um ganz genau zu sein: Die `trusted list' ist eine Datei von Zeilen,
**   die (neben Kommentaren und Leerzeilen) Zeilen mit absoluten Pfadnamen
**   von Programmen enthlt, die von `vxdeliver' ausgefhrt werden drfen.
**   Diese Liste wird - wie auch die Default-Konfigurationsdatei unter
**   `/etc' stehen; zwar kann irgendjemand durch Neubersetzung von
**   `vxdeliver' seine eigene `trusted list' erzeugen; jedoch wird niemals
**   die Mglichkeit bestehen, da derjenige etwa `root'-Zugangsrechte
**   erlangt, weil seine Version von `vxdeliver' nicht mit diesen Rechten
**   ausgestattet ist.
*/
#define TRUSTED "/etc/vxmail.progs"

struct trusted_s;

typedef struct trusted_s *trusted_t;

struct trusted_s {
    trusted_t next;
    char *prog;
    int mf;
    uid_t uid;
    gid_t gid;
};

static trusted_t trusted_list = (trusted_t) 0;

/* Fge einen Eintrag in die `trusted list' ein. */
static int add_trusted (const char *path, int lc)
{
    char ugbf[1024], *p = (char *) path, *q;
    int mf = 0;
    uid_t tuid = prefs.vxuid;
    gid_t tgid = prefs.vxgid;
    struct passwd *pw;
    struct group *gr;
    trusted_t tle;
    if  (*p == '(') {
	/* Wenn die angegebene Zeile mit (uname:gname) beginnt, dann
	** bekommt der Eintrag in der `trusted list' die angegebene
	** User- bzw. Group-ID. Das jeweilige Programm wird dann spter
	** mit genau diesen ausgefhrt. Da die nderung jedoch nur dann
	** geschehen kann, wenn `vxdeliver' mit `root'-Rechten luft, sollte
	** keinerlei Sicherheitsrisiko damit verbunden sein.
	*/
	++p; q = strchr (p, ':');
	if (!q) {
	    vxlog (0, "ERROR: missing group id (%s:%d)!\n", lc, TRUSTED);
	    return 0;
	}
	strncpy (ugbf, p, (q - p)); ugbf[(q - p)] = '\0';
	if (!(pw = getpwnam (ugbf))) {
	    vxlog (0, "ERROR: unknown user `%s' (%s:%d)!\n", ugbf, TRUSTED, lc);
	    return 0;
	}
	tuid = pw->pw_uid;
	p = q + 1;
	q = strchr (p, ')');
	if (!q) {
	    vxlog (0, "ERROR: missing pathname (%s:%d)!\n", lc, TRUSTED);
	    return 0;
	}
	strncpy (ugbf, p, (q - p)); ugbf[(q - p)] = '\0';
	if (!(gr = getgrnam (ugbf))) {
	    vxlog (0, "ERROR: unknown group `%s' (%s:%d)!\n",
		      ugbf, TRUSTED, lc);
	    return 0;
	}
	tgid = gr->gr_gid;
	p = q + 1;
	while (*p && strchr (" \t", *p)) { ++p; }
	if (!*p) {
	    vxlog (0, "ERROR: missing pathname (%s:%d)!\n", lc, TRUSTED);
	    return 0;
	}
    }
    if (*p == 'F' || *p == 'f') {
	mf = 1;
	++p; while (*p && strchr (" \t", *p)) { ++p; }
	if (!*p) {
	    vxlog (0, "ERROR: missing pathname (%s:%d)!\n", lc, TRUSTED);
	    return 0;
	}
    }
    if (*p != '/') {
	vxlog (0, "ERROR: pathname must be absolute (%s:%d)!\n", lc, TRUSTED);
	return 0;
    }
    tle = (trusted_t) galloc (__FILE__, __LINE__,
			      sizeof(struct trusted_s) + strlen (p) + 1);
    tle->mf = mf;
    tle->uid = tuid;
    tle->gid = tgid;
    tle->prog = (char *) tle + sizeof (struct trusted_s);
    strcpy (tle->prog, p);
    tle->next = trusted_list;
    trusted_list = tle;
    vxlog (5, "    added `%s' (%d,%d).\n", tle->prog, tle->uid, tle->gid);
    return 1;
}

int read_trusted (void)
{
    FILE *tlfp;
    char line[1024], *p;
    int cc = 0, lc = 0, lognl = 1;
    size_t ll;
    if (!(tlfp = fopen (TRUSTED, "r"))) { return cc; }
    while (fgets (line, 1024, tlfp)) {
	++lc; ll = strlen (line);
	if (line[ll - 1] != '\n') {
	    vxlog (0, "ERROR: line %d in file `%s' is too long!\n",
		      lc, TRUSTED);
	    do {
		if (!fgets (line, 1024, tlfp)) { break; }
		ll = strlen (line);
	    } while (line[ll - 1] != '\n');
	} else {
	    line[--ll] = '\0';
	    for (p = line; *p && strchr (" \t", *p); ++p);
	    if (!*p || *p == '#') { continue; }
	    if (lognl) { vxlog (5, "\n"); lognl = 0; }
	    if (add_trusted (p, lc)) { ++cc; }
	}
    }
    fclose (tlfp);
    return cc;
}

/* Prfe, ob das durch (bergebene) Kommandozeile spezifizierte Programm
** ausgefhrt werden darf oder nicht. Gib im Erfolgsfall die entsprechend
** modifizierte Kommandozeile als Ergebnis und die erwartete UID/GID ber
** die zustzlichen Argumente zurck; im Fehlerfall wird NULL zurckgegeben
** und die zustzlichen Argumente werden nicht verndert.
*/
char *valid_prog (char *cmdline, uid_t *tuid, gid_t *tgid, int mf)
{
    char *p, *prog, *newcmd;
    trusted_t tle;
    /* check if executing commands is generally allowed. */
    if (!prefs.rexecdir) { return NULL; }

    /* Skip leading spaces */
    while (*cmdline == ' ' || *cmdline == '\t') { ++cmdline; }

    /* search for the next SPACE or TAB character in `cmdline'. */
    if (!(p = strpbrk (cmdline, " \t"))) { p = cmdline + strlen (cmdline); }

    /* allocate a new memory area for the program's pathname */
    prog = galloc (__FILE__, __LINE__, (p - cmdline) + 1);
    /* ... and fill it with the program's pathname */
    memcpy (prog, cmdline, (p - cmdline));
    prog[(p - cmdline)] = '\0';

    /* Check if the program is in the trusted list */
    for (tle = trusted_list; noNULL(tle); tle = tle->next) {
	size_t ol = strlen (tle->prog), cl = strlen (prog);
	if (ol >= cl && tle->mf == mf && !strcmp(tle->prog+ol-cl, prog)) {
	    break;
	}
    }
    ifnull (tle) { gfree (prog); return NULL; }

    *tuid = tle->uid;
    *tgid = tle->gid;

    /* if the program's pathname is VALID, generate a new command line
    ** from it and the rest of the supplied command line.
    */
    newcmd = gconc (__FILE__, __LINE__, tle->prog, p, NULL);
    gfree (prog);
    return newcmd;
}
